import asyncio
import logging
import pathlib
import random

import discord

from botbotbot.ai import AIBot
from botbotbot.shuffle import Shuffler
from botbotbot.tts import CambAI

logger = logging.getLogger(__name__)


class VoiceBot:
    def __init__(
        self,
        bot: discord.Bot,
        cambai: CambAI | None = None,
        aibot: AIBot | None = None,
        shuffler: Shuffler | None = None,
        guild_ids: list[int] = [],
    ) -> None:
        self.bot = bot
        self.cambai = cambai
        self.shf = shuffler
        self.aibot = aibot
        self.guild_ids = guild_ids
        self.vo: dict[int, discord.VoiceClient] = dict()

        self.mots = pathlib.Path("assets/mots.txt").read_text().split("\n")

    def init_events(self) -> None:
        self.bot.add_listener(self.on_voice_state_update, "on_voice_state_update")
        self.bot.add_listener(self.shuffle, "on_voice_state_update")
        self.bot.add_listener(self.random_connect, "on_ready")
        self.bot.add_application_command(
            discord.SlashCommand(
                self.join_voice,
                name="join",
                description="Rejoindre un channel vocal",
                guild_ids=self.guild_ids,
            )
        )
        self.bot.add_application_command(
            discord.SlashCommand(
                self.unjoin_voice,
                name="unjoin",
                description="Déconnecter du channel vocal",
                guild_ids=self.guild_ids,
            )
        )

    async def random_connect(self) -> None:
        while True:
            for g_id in self.guild_ids:
                guild = self.bot.get_guild(g_id)
                if guild is None:
                    continue

                mot = random.choice(self.mots)
                if self.aibot is not None:
                    status = self.aibot.answer(
                        f"Dis une phrase aléatoire et originale avec le mot {mot}."
                    )
                    activity = discord.CustomActivity(status)
                    await self.bot.change_presence(activity=activity)

                weights = [10 * len(vc.members) + 1 for vc in guild.voice_channels]
                if random.random() < sum(weights) / 400:
                    logger.info("Random connect.")
                    voice_chan = random.choices(
                        guild.voice_channels,
                        weights=weights,
                    )[0]
                    await self.connect_voice(voice_chan)
                    if not voice_chan.status and self.aibot is not None:
                        mot = random.choice(self.mots)
                        await voice_chan.set_status(
                            self.aibot.answer(
                                f"Dis une phrase aléatoire et originale avec le mot {mot}."
                            )
                        )
                else:
                    logger.info("Random disconnect.")
                    await self.disconnect_voice(guild)

            await asyncio.sleep(3600 * random.random())

    async def connect_voice(
        self, channel: discord.VoiceChannel | discord.StageChannel
    ) -> None:
        if self.cambai is None:
            return

        logger.info(f"Connecting to voice channel <{channel}>.")
        if len(channel.members) == 0:
            await self.connect_voice_chan(channel)
            return
        elif len(channel.members) == 1:
            member = self.rnd_name(channel.members[0])
        elif random.random() < 50 / 100:
            member = self.rnd_name(random.choice(channel.members))
        else:
            member = (
                ", ".join(self.rnd_name(member) for member in channel.members[:-1])
                + " et "
                + self.rnd_name(channel.members[-1])
            )

        script = None
        if self.aibot is not None and random.random() < 20 / 100:
            mot = random.choice(self.mots)
            script = self.aibot.answer(
                f"Dis une phrase aléatoire et originale avec le mot {mot}."
            )
        if script is None:
            script = random.choice(
                [
                    "Salut la jeunesse !",
                    f"Salut {member}, ça va bien ?",
                    "Allo ? À l'huile !",
                ]
            )
        source = await discord.FFmpegOpusAudio.from_probe(self.cambai.tts(script))
        vo = await self.connect_voice_chan(channel)
        await asyncio.sleep(10 * random.random())
        await vo.play(source, wait_finish=True)

    @classmethod
    def rnd_name(cls, user: discord.Member) -> str:
        return random.choice([user.display_name, user.name])

    async def connect_voice_chan(
        self, channel: discord.VoiceChannel | discord.StageChannel
    ) -> discord.VoiceClient:
        vo = self.vo.get(channel.guild.id, None)
        if vo is not None and vo.channel != channel:
            await self.disconnect_voice(channel.guild)
        if vo is None or vo.channel != channel:
            vo = await channel.connect()
        self.vo[channel.guild.id] = vo
        return vo

    async def disconnect_voice(self, guild: discord.Guild) -> None:
        if vo := self.vo.pop(guild.id, None):
            if self.cambai is not None:
                source = await discord.FFmpegOpusAudio.from_probe(
                    self.cambai.tts("À plus les jeunes !")
                )
                await vo.play(source, wait_finish=True)
            await vo.disconnect()

    async def join_voice(self, ctx: discord.ApplicationContext) -> None:
        if ctx.user.voice:
            await ctx.respond(
                f"Connecting to voice channel {ctx.user.voice.channel}.",
                ephemeral=True,
                delete_after=30,
            )
            await self.connect_voice(ctx.user.voice.channel)
        else:
            await ctx.respond(
                "You are not connected to a voice channel.",
                ephemeral=True,
                delete_after=30,
            )

    async def unjoin_voice(self, ctx: discord.ApplicationContext) -> None:
        if ctx.user.voice:
            await ctx.respond(
                f"Disconnecting from voice channel {ctx.user.voice.channel}.",
                ephemeral=True,
                delete_after=30,
            )
            await self.disconnect_voice(ctx.user.voice.channel.guild)
        else:
            await ctx.respond(
                "You are not connected to a voice channel.",
                ephemeral=True,
                delete_after=30,
            )

    async def on_voice_state_update(
        self,
        member: discord.Member,
        before: discord.VoiceState,
        after: discord.VoiceState,
    ) -> None:
        logger.debug("Voice state update")
        logger.debug(before.channel)
        logger.debug(after.channel)
        if after.channel:
            logger.debug(after.channel.members)

        if (
            self.cambai is None
            or after.channel is None
            or before.channel == after.channel
            or (self.bot.user and member == self.bot.user)
        ):
            return

        if self.bot.user in after.channel.members and (
            vo := self.vo.get(after.channel.guild.id, None)
        ):
            m = self.rnd_name(member)
            script = f"Salut {m}, ça va bien ?"
            source = await discord.FFmpegOpusAudio.from_probe(self.cambai.tts(script))
            await asyncio.sleep(10 * random.random())
            await vo.play(source, wait_finish=True)
            return

        if random.random() < 5 / 100:
            await self.connect_voice(after.channel)

    async def shuffle(
        self,
        member: discord.Member,
        before: discord.VoiceState,
        after: discord.VoiceState,
    ) -> None:
        if (
            self.shf
            and before.channel is None
            and member != self.bot.user
            and random.random() < 5 / 100
        ):
            logger.info(f"Voice shuffle from {member}")
            await self.shf.try_shuffle(member.guild)
